#ifndef DENSE_H
#define DENSE_H

#include "kktdirect.h"

// Dense kernels from BLAS and LAPACK for supernodal mode
// Make sure you #define one of USE_FORTRAN_BLAS or USE_CBLAS,
// and that you #define USE_FORTRAN_LAPACK (currently there is no support for e.g. ATLAS clapack_* calls)

/*************************************************************************************************************
The wrapped BLAS calls are based on DGEMM:
 
void BLAS::general_matrix_matrix_multiply(KKT_INT m, KKT_INT n, KKT_INT k, const double *A, const double *B, double *C);
Multiply m*k matrix A with k*n matrix B, both stored column-wise, and store m*n result in C column-wise.

void BLAS::general_matrix_matrix_multiply(CBLAS_TRANSPOSE transposeA, CBLAS_TRANSPOSE transposeB,
                                          KKT_INT m, KKT_INT n, KKT_INT k, const double *A, const double *B, double *C);
Multiply A or A^T with B or B^T, both stored column-wise, and store m*n result in C column-wise.
Note that if A is not transposed, its dimensions are m*k; if it is transposed, the dimensions of A are k*m.
Similarly if B is not transposed, its dimensions are k*n; if it is transposed, the dimensions of B are n*k.

void BLAS::general_matrix_matrix_multiply(CBLAS_TRANSPOSE transposeA, CBLAS_TRANSPOSE transposeB,
                                          KKT_INT m, KKT_INT n, KKT_INT k, double alpha, const double *A, const double *B,
                                          double beta, double *C);
Multiply alpha times A or A^T with B or B^T, both stored column-wise, and add m*n result to beta times C
stored column-wise.
Note that if A is not transposed, its dimensions are m*k; if it is transposed, the dimensions of A are k*m.
Similarly if B is not transposed, its dimensions are k*n; if it is transposed, the dimensions of B are n*k.
Constants for CBLAS_TRANPOSE are either CblasTrans or CblasNoTrans.
*************************************************************************************************************/

#ifdef USE_FORTRAN_BLAS
//============================================ FORTRAN BLAS ==================================================
// I assume lowercase with an underscore suffix convention; if your platform does things differently you will
// need to modify this!

typedef const char *CBLAS_TRANSPOSE;
#define CblasNoTrans "N"
#define CblasTrans "T"

extern "C" {
void dgemm_(const char *TRANSA, const char *TRANSB, const KKT_INT *M, const KKT_INT *N, const KKT_INT *K,
            const double *ALPHA, const double *A, const KKT_INT *LDA, const double *B, const KKT_INT *LDB,
            const double *BETA, double *C, const KKT_INT *LDC);
}

namespace BLAS{
   inline void general_matrix_matrix_multiply(KKT_INT m, KKT_INT n, KKT_INT k, const double *A, const double *B, double *C)
   {
      const double one=1, zero=0;
      dgemm_(CblasNoTrans, CblasNoTrans, &m, &n, &k, &one, A, &m, B, &k, &zero, C, &m);
   }
   
   inline void general_matrix_matrix_multiply(CBLAS_TRANSPOSE transposeA, CBLAS_TRANSPOSE transposeB,
                                              KKT_INT m, KKT_INT n, KKT_INT k, const double *A, const double *B, double *C)
   {
      const double one=1, zero=0;
      dgemm_(transposeA, transposeB, &m, &n, &k, &one,
                         A, (transposeA==CblasNoTrans ? &m : &k),
                         B, (transposeB==CblasNoTrans ? &k : &n),
                         &zero, C, &m);
   }

   inline void general_matrix_matrix_multiply(CBLAS_TRANSPOSE transposeA, CBLAS_TRANSPOSE transposeB,
                                              KKT_INT m, KKT_INT n, KKT_INT k, double alpha, const double *A, const double *B,
                                              double beta, double *C)
   {
      dgemm_(transposeA, transposeB, &m, &n, &k, &alpha,
             A, (transposeA==CblasNoTrans ? &m : &k),
             B, (transposeB==CblasNoTrans ? &k : &n),
             &beta, C, &m);
   }
};

#elif defined USE_CBLAS
//=============================================== cblas ======================================================
#include "cblas.h"

namespace BLAS{
   inline void general_matrix_matrix_multiply(KKT_INT m, KKT_INT n, KKT_INT k, const double *A, const double *B, double *C)
   {
      cblas_dgemm(CblasColMajor, CblasNoTrans, CblasNoTrans, m, n, k, 1.0, A, m, B, k, 0.0, C, m);
   }
   
   inline void general_matrix_matrix_multiply(CBLAS_TRANSPOSE transposeA, CBLAS_TRANSPOSE transposeB,
                                              KKT_INT m, KKT_INT n, KKT_INT k, const double *A, const double *B, double *C)
   {
      cblas_dgemm(CblasColMajor, transposeA, transposeB, m, n, k, 1.0,
                  A, (transposeA==CblasNoTrans ? m : k),
                  B, (transposeB==CblasNoTrans ? k : n),
                  0.0, C, m);
   }

   inline void general_matrix_matrix_multiply(CBLAS_TRANSPOSE transposeA, CBLAS_TRANSPOSE transposeB,
                                              KKT_INT m, KKT_INT n, KKT_INT k, double alpha, const double *A, const double *B,
                                              double beta, double *C)
   {
      cblas_dgemm(CblasColMajor, transposeA, transposeB, m, n, k, alpha,
                  A, (transposeA==CblasNoTrans ? m : k),
                  B, (transposeB==CblasNoTrans ? k : n),
                  beta, C, m);
   }
};

#endif

/**************************************************************************************************************
The wrapped LAPACK calls are for DPOTRF, DPOTRS, DSYTRF and DSYTRS. In addition, I provide a routine which uses
DPOTRF/S under the hood for a non-pivoting version of DSYTRF/S (when supplied information on positive vs. negative
defineness).

void LAPACK::factor_symmetric_positive_definite(KKT_INT n, double *A, KKT_INT &info);
Factor symmetric positive definite n*n matrix A in-place with Cholesky.
info is set on output to indicate if factorization succeeded (0).

void LAPACK::solve_symmetric_positive_definite_with_factorization(KKT_INT n, KKT_INT nrhs,
                                                                  const double *factorization,
                                                                  double *righthandside, KKT_INT &info);
Given the output from factor_symmetric_positive_definite, solve nrhs linear systems with given
righthandside (storing result in-place)


void LAPACK::factor_symmetric_indefinite(KKT_INT n, double *A, KKT_INT *ipiv, double *work, KKT_INT worksize,
                                         KKT_INT &info);
Factor symmetric indefinite n*n matrix A in-place with Bunch-Kaufmann.
ipiv should be a length n vector (on output: the pivot sequence used with signs indicating 1x1 or 2x2 blocks), 
work should be a vector of length worksize, used internally (ideally at least of size n*internal_blocksize)
info is set on output to indicate if factorization succeeded (0).

void LAPACK::solve_symmetric_indefinite_with_factorization(KKT_INT n, KKT_INT nrhs,
                                                           const double *factorization, const KKT_INT *ipiv,
                                                           double *righthandside, KKT_INT &info);
Given the output from factor_symmetric_indefinite, solve nrhs linear systems with given righthandside (storing
result in-place)


void LAPACK::factor_symmetric_indefinite(KKT_INT n, double *A, const char *constrained, KKT_INT &info);
void LAPACK::solve_symmetric_indefinite_with_factorization(KKT_INT n, KKT_INT nrhs,
                                                           const double *factorization, const char *constrained,
                                                           double *righthandside, KKT_INT &info);

*************************************************************************************************************/

#ifdef USE_FORTRAN_LAPACK
//========================================== FORTRAN LAPACK ================================================
// Note: should be compatible with f2c'd CLAPACK as well (just don't bother including the clapack.h header file)
// and Apple's vecLib LAPACK.

extern "C" {
void dpotrf_(const char *UPLO, KKT_INT *N, double *A, KKT_INT *LDA, KKT_INT *INFO);
void dpotrs_(const char *UPLO, KKT_INT *N, KKT_INT *NRHS, double *A, KKT_INT *LDA,
             double *B, KKT_INT *LDB, KKT_INT *INFO);
void dsytrf_(const char *UPLO, KKT_INT *N, double *A, KKT_INT *LDA, KKT_INT *IPIV,
             double *WORK, KKT_INT *LWORK, KKT_INT *INFO);
void dsytrs_(const char *UPLO, KKT_INT *N, KKT_INT *NRHS, double *A, KKT_INT *LDA, KKT_INT *IPIV,
             double *B, KKT_INT *LDB, KKT_INT *INFO);
}

namespace LAPACK{
   inline void factor_symmetric_positive_definite(KKT_INT n, double *A, KKT_INT &info)
   {
      dpotrf_("L", &n, A, &n, &info);
   }

   inline void solve_symmetric_positive_definite_with_factorization(KKT_INT n, KKT_INT nrhs,
                                                                    const double *factorization,
                                                                    double *righthandside, KKT_INT &info)
   {
      dpotrs_("L", &n, &nrhs, (double*)factorization, &n, righthandside, &n, &info);
   }

   inline void factor_symmetric_indefinite(KKT_INT n, double *A, KKT_INT *ipiv, double *work, KKT_INT worksize, KKT_INT &info)
   {
      dsytrf_("L", &n, A, &n, ipiv, work, &worksize, &info);
   }
   
   inline void solve_symmetric_indefinite_with_factorization(KKT_INT n, KKT_INT nrhs, const double *factorization, const KKT_INT *ipiv,
                                                             double *righthandside, KKT_INT &info)
   {
      dsytrs_("L", &n, &nrhs, (double*)factorization, &n, (KKT_INT*)ipiv, righthandside, &n, &info);
   }
};

#endif

#endif
